/*************************************************************************\
* Copyright (c) 2000-2004 The University of Chicago, as Operator of Argonne
* National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
// Methods for CHist class

#define DEBUG_CA 0

// Unreferenced
#ifdef __cplusplus
#define UNREFERENCED(x) (x=x)
#else
#define UNREFERENCED(x)
#endif

#define HIST_DIR_SDDS_MASK "*.sdds"
#define HIST_DIR_EPS_MASK "*.eps"

#define QUIT_BTN              0
#define TOOLS_START_BTN       1
#define TOOLS_STOP_BTN        2
#define TOOLS_RESET_BTN       3
#define TOOLS_OPTIONS_BTN     4
#define OPTIONS_APPLY_BTN     5
#define OPTIONS_OK_BTN        6
#define OPTIONS_CANCEL_BTN    7
#define OPTIONS_SAVE_BTN      8
#define OPTIONS_HELP_BTN      9
#define FILE_SAVE_BTN         10
#define FILE_SAVEAS_SDDS_BTN  11
#define FILE_SAVEAS_EPS_BTN   12

#define OPTIONS_INDENT 25

// These must start with 3 or greater, 0-2 are for option menu buttons
#define OPTIONS_BIN_MODE_BTN            101
#define OPTIONS_BIN_LOW_ENTRY           102
#define OPTIONS_BIN_HIGH_ENTRY          103
#define OPTIONS_BIN_NBINS_ENTRY         104
#define OPTIONS_BIN_UPDATE_PERIOD_ENTRY 105
#define OPTIONS_AXIS_MODE_BTN           106
#define OPTIONS_AXIS_MIN_ENTRY          107
#define OPTIONS_AXIS_MAX_ENTRY          108

#define HIST_DEFAULT_HEIGHT 400
#define HIST_DEFAULT_WIDTH  400
#define MARGIN_HEIGHT 10
#define MARGIN_WIDTH 10
#define PLOT_X 10
#define PLOT_Y 10
#define TITLE_FONT_HEIGHT 15
#define AXIS_FONT_HEIGHT 10
#define LABEL_FONT_HEIGHT 10
#define MIN_FONT_HEIGHT 8

#define MAX(x,y) (((x)>=(y))?(x):(y))
#define MIN(x,y) (((x)<=(y))?(x):(y))

#define ERRMSG if(errHandler) (*errHandler)

#ifdef WIN32
/* WIN32 does not have unistd.h and does not define the following constants */
#define F_OK 00
#define W_OK 02
#define R_OK 04
#include <direct.h>     /* for getcwd (usually in sys/parm.h or unistd.h) */
#include <io.h>         /* for access, chmod  (usually in unistd.h) */
#else
/* Use unistd.h */
#include <unistd.h>
#endif
#include <errno.h>

#include <Xm/Xm.h>
#include <Xm/CascadeB.h>
#include <Xm/FileSB.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/MainW.h>
#include <Xm/PanedW.h>
#include <Xm/PushB.h>
#include <Xm/RowColumn.h>
#include <Xm/Separator.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>

#include <Xm/MwmUtil.h>
#include <Xm/Protocols.h>
#include <X11/Xatom.h>
#include "SciPlot.h"
#include "histTool.h"

// Function prototypes

static void valueChangedCB(CCaObject2 *chanIn);
static void graphicalInfoCB(CCaObject2 *chanIn);
static void pollTimerProc(XtPointer cd, XtIntervalId *id);
static void histTimerProc(XtPointer cd, XtIntervalId *id);

// Global variables

static Atom WM_DELETE_WINDOW;
static XtIntervalId pollTimerId=0;
static Widget optNameW,optBinModeW,optBinLowE,optBinHighE;
static Widget optBinNBinsE,optBinUpdatePeriodE,optBinLoprE,optBinHoprE;
static Widget optAxisModeW,optAxisMinE,optAxisMaxE;
static Widget optionsApplyB,optionsOkB,optionsCancelB,optionsSaveB,optionsHelpB;
static int fsdButton;
static Widget fileSD=NULL;
static Widget optS=NULL;

// Static initializations

CHistErrorHandler CHist::errHandler=errMsg;
CHistHelpCallback CHist::helpCB=NULL;
unsigned long CHist::refCount=0;
unsigned long CHist::nextId=0;

CHist::CHist(char *pvNameIn) :
    binMode(options.binMode),
    binUserMin(options.binUserMin),
    binUserMax(options.binUserMax),
    binUpdatePeriod(options.binUpdatePeriod),
    nBins(options.nBins),
    axisMode(options.axisMode),
    axisXMin(options.axisXMin),
    axisXMax(options.axisXMax),
    nPoints(0),
    binMin(0.0),
    binMax(0.0),
    binWidth(0.0),
    binTotalWidth(0.0),
    value(0.0),
    initialized(0),
    chan(NULL),
    needsUpdate(1),
    histTimerId(0),
    minY(0.),
    maxY(1.),
    xArray(NULL),
    yArray(NULL),
    yAxisStyle(LINEAR_AXIS),
    yRangeStyle(AUTO_SCALE_RANGE),
    plotID(0),
    hopr(0.0),
    lopr(0.0),
    prec(2),
    shell(NULL),
    mainForm(NULL),
    plotW(NULL)
{
  // Use the next id number
    id=nextId++;

  // PV name
    strncpy(pvName,pvNameIn,CA_NAMESIZE);
    pvName[CA_NAMESIZE-1]='\0';

  // Filename
    *fileName='\0';
    fileType=SDDS_File;
    
  // Allocate caObject
    chan=new CCaObject2;
    if(!chan) {
	ERRMSG("Cannot allocate a CCaObject2");
    }
    chan->setUserData(this);
    chan->setGraphicalInfoCB(::graphicalInfoCB);
    chan->setValueChangedCB(::valueChangedCB);
    chan->setErrorHandler(errMsg);
    chan->setMonitorValueChanged(1);
    if(*pvName) {
	chan->connect(pvName);
	chan->startMonitor();
    }

  // Start a timer proc for ca_poll
    refCount++;
    if(!pollTimerId) {
      // Initialize these static variables the first time
	pollTimerId=XtAppAddTimeOut(appContext,POLL_INTERVAL,::pollTimerProc,NULL);
    }

  // Other initialization
    initialize();
}

CHist::~CHist(void)
{
  // Popdown the options dialog if it is set for this instance
    if(optS) {
	CHist *hist=NULL;
	XtVaGetValues(optionsOkB,XmNuserData,&hist,NULL);
	if(hist == NULL || hist == this) {
	    XtPopdown(optS);
	}
    }
    if(shell) {
	XtPopdown(shell);
	XtDestroyWidget(shell);
    }
    if(windowList) {
	windowList->remove(*this);
    }
    if(xArray) delete [] xArray;
    if(yArray) delete [] yArray;

  // Stop the histogram timer
    if(histTimerId) {
	XtRemoveTimeOut(histTimerId);
	histTimerId=(XtIntervalId)0;
    }

  // Delete the chan
    if(chan) delete chan;    

  // Stop the poll timer proc if no more instantiations left
    refCount--;
    if(refCount <= 0 && pollTimerId) {
	XtRemoveTimeOut(pollTimerId);
	pollTimerId=(XtIntervalId)0;
    }
}

int CHist::writeSDDSFile(const char * fileNameIn, Boolean check)
{
    FILE *fp=NULL;
    int writeOK=0;
    int i;

  // Check if the file exists
    if(check) {
	errno=0;
	if(access(fileNameIn,W_OK) == -1) {
	    if(errno == ENOENT) {
	      // File does not exist
		writeOK=1;
	    } else {
	      // There is some error
		char *errstring=strerror(errno);
		ERRMSG("Error accessing file:\n%s\n%s\n",
		  fileNameIn,errstring);
		return HT_ERROR;
	    }
	} else {
	    char string[80];
	    
	  // File exists, see whether the user wants to overwrite
	    sprintf(string,"Do you want to overwrite file:\n%s",
	      fileNameIn);
	    int answer=yesNo(string,NULL,NULL);
	    if(answer == 1) writeOK=1;
	}
	if(!writeOK) return 0;
    }

  // Try to open the file
    fp=fopen(fileNameIn,"w");
    if(fp == NULL) {
	errMsg("Cannot open file: %s\n",fileNameIn);
	goto CLEANUP;
    }
    
  // Write the header
    fprintf(fp,"SDDS1\n");
    fprintf(fp,"&description text=\"HistTool Output File\"\n");
    fprintf(fp,"  contents=\"HistTool Output File\" &end\n");
    fprintf(fp,"&parameter name=pvName type=string &end\n");
    fprintf(fp,"&parameter name=binMin type=double &end\n");
    fprintf(fp,"&parameter name=binMax type=double &end\n");
    fprintf(fp,"&parameter name=nBins type=long &end\n");
    fprintf(fp,"&parameter name=binUpdatePeriod type=long &end\n");
    fprintf(fp,"&column name=%s type=double &end\n", pvName?pvName:"Null");
    fprintf(fp,"&column name=Bin type=long &end\n");
    fprintf(fp,"&column name=Count type=double &end\n");
    fprintf(fp,"&data mode=ascii no_row_counts=1 additional_header_lines=1 &end\n");
    fprintf(fp,"\n");
    fprintf(fp,"%s\n",pvName?pvName:"Null");
    fprintf(fp,"%.15g\n",binMin);
    fprintf(fp,"%.15g\n",binMax);
    fprintf(fp,"%d\n",nBins);
    fprintf(fp,"%u\n",binUpdatePeriod);
    
  // Write the bins
    for(i=0; i < nBins; i++) {
	fprintf(fp,"%21.15f %3d %21.15f\n",binMin+i*(binMax-binMin)/(nBins-1), i+1, yArray[2*i+1]);
    }
    
  // Close the file
    fclose(fp);
    strncpy(fileName,fileNameIn,PATH_MAX);
    fileName[PATH_MAX-1]='\0';
    fileType=SDDS_File;
    return 1;

  // Clean up
  CLEANUP:
    if(fp) fclose(fp);
    return 0;
}

int CHist::writeEPSFile(const char * fileNameIn, Boolean check)
{
    FILE *fp=NULL;
    int writeOK=0;
    Boolean status;

  // Check if the file exists
    if(check) {
	errno=0;
	if(access(fileNameIn,W_OK) == -1) {
	    if(errno == ENOENT) {
	      // File does not exist
		writeOK=1;
	    } else {
	      // There is some error
		char *errstring=strerror(errno);
		ERRMSG("Error accessing file:\n%s\n%s\n",
		  fileNameIn,errstring);
		return HT_ERROR;
	    }
	} else {
	    char string[80];
	    
	  // File exists, see whether the user wants to overwrite
	    sprintf(string,"Do you want to overwrite file:\n%s",
	      fileNameIn);
	    int answer=yesNo(string,NULL,NULL);
	    if(answer == 1) writeOK=1;
	}
	if(!writeOK) return 0;
    }

  // Try to open the file
    fp=fopen(fileNameIn,"w");
    if(fp == NULL) {
	errMsg("Cannot open file: %s\n",fileNameIn);
	return 0;
    }
    fclose(fp);

  // Make a non-const filename for SciPlot
    char tempName[PATH_MAX];
    strncpy(tempName,fileNameIn,PATH_MAX);
    tempName[PATH_MAX-1]='\0';

  // Write the file
    status=SciPlotPSCreate(plotW,tempName);
    if(status) {
	strncpy(fileName,fileNameIn,PATH_MAX);
	fileName[PATH_MAX-1]='\0';
	fileType=EPS_File;
	return 1;
    } else {
	return 0;
    }
}

void CHist::initialize(void)
{
    int first=1;
    Widget w,w1,wparent;
    Widget mainWindow=NULL;
    Widget mainMenu=NULL;

  // Initialize static stuff
    if(first) {
	WM_DELETE_WINDOW=XmInternAtom(display,"WM_DELETE_WINDOW",False);
	first=0;
    }

  // Allocate colors !!!!!!!!!!!!!!!!!!!!
    bgColor=WhitePixel(display,DefaultScreen(display));
    fgColor=BlackPixel(display,DefaultScreen(display));

  // If the shell is created, just pop it up
    if(shell) {
	XtPopup(shell,XtGrabNone); 
	return;
    }

  // Create shell if not created
    nargs = 0;
#if 0
  // The following should turn all decorations on except resize handles
    XtSetArg(args[nargs],XmNmwmDecorations,
      MWM_DECOR_ALL|MWM_DECOR_RESIZEH); nargs++;
    XtSetArg(args[nargs],XmNallowShellResize,TRUE); nargs++;

  // This should not be necessary but it fixes Exceed problem
    XtSetArg(args[nargs],XmNmwmFunctions,
      MWM_FUNC_ALL); nargs++;
#endif

  // Map window manager menu Close function to nothing for now
    XtSetArg(args[nargs],XmNdeleteResponse,XmDO_NOTHING); nargs++;
    XtSetArg(args[nargs],XmNtitle,pvName); nargs++;
    shell = XtCreatePopupShell("plotShell",topLevelShellWidgetClass,
      appMainWindow,args,nargs);
    
  // Fix up WM close
    XmAddWMProtocolCallback(shell,WM_DELETE_WINDOW,
      quit,(XtPointer)this);
    
  /* Define main window */
    nargs=0;
  // KE: Should be done with fallback resources !!!!!
    XtSetArg(args[nargs],XmNshadowThickness,0); nargs++;
    mainWindow=XmCreateMainWindow(shell,"plotWindow",args,nargs);
    XtManageChild(mainWindow);

  /* Define menu bar */
    nargs=0;
    mainMenu=XmCreateMenuBar(mainWindow,"mainMenu",args,nargs);
    XtManageChild(mainMenu);
    wparent=mainMenu;
    
  /* Define file menu */
    nargs=0;
    XtSetArg(args[nargs],XmNtearOffModel,XmTEAR_OFF_ENABLED); nargs++;
    w1=XmCreatePulldownMenu(mainMenu,"fileMenu",args,nargs);
    
    nargs=0;
    XtSetArg(args[nargs],XmNsubMenuId,w1); nargs++;
    w=XmCreateCascadeButton(mainMenu,"File",args,nargs);
    XtManageChild(w);
    
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,this); nargs++;
    w=XmCreatePushButton(w1,"Save",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,(XtPointer)FILE_SAVE_BTN);

    nargs=0;
    XtSetArg(args[nargs],XmNuserData,this); nargs++;
    w=XmCreatePushButton(w1,"Save As SDDS...",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,
      (XtPointer)FILE_SAVEAS_SDDS_BTN);

    nargs=0;
    XtSetArg(args[nargs],XmNuserData,this); nargs++;
    w=XmCreatePushButton(w1,"Save As EPS...",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,
      (XtPointer)FILE_SAVEAS_EPS_BTN);

    nargs=0;
    XtSetArg(args[nargs],XmNuserData,this); nargs++;
    w=XmCreatePushButton(w1,"Quit",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,(XtPointer)QUIT_BTN);
    
  /* Define tools menu */
    nargs=0;
    XtSetArg(args[nargs],XmNtearOffModel,XmTEAR_OFF_ENABLED); nargs++;
    w1=XmCreatePulldownMenu(mainMenu,"toolsMenu",args,nargs);
    
    nargs=0;
    XtSetArg(args[nargs],XmNsubMenuId,w1); nargs++;
    w=XmCreateCascadeButton(mainMenu,"Tools",args,nargs);
    XtManageChild(w);
    
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,this); nargs++;
    w=XmCreatePushButton(w1,"Start",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,(XtPointer)TOOLS_START_BTN);
    
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,this); nargs++;
    w=XmCreatePushButton(w1,"Stop",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,(XtPointer)TOOLS_STOP_BTN);
    
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,this); nargs++;
    w=XmCreatePushButton(w1,"Reset",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,(XtPointer)TOOLS_RESET_BTN);

    nargs=0;
    XtSetArg(args[nargs],XmNuserData,this); nargs++;
    w=XmCreatePushButton(w1,"Options...",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,mainCB,(XtPointer)TOOLS_OPTIONS_BTN);
    
  /* Define help menu */
    nargs=0;
    XtSetArg(args[nargs],XmNtearOffModel,XmTEAR_OFF_ENABLED); nargs++;
    w1=XmCreatePulldownMenu(mainMenu,"helpMenu",args,nargs);
    
    nargs=0;
    XtSetArg(args[nargs],XmNsubMenuId,w1); nargs++;
    w=XmCreateCascadeButton(mainMenu,"Help",args,nargs);
    XtManageChild(w);
    
    nargs=0;
    XtSetArg(args[nargs],XmNmenuHelpWidget,w); nargs++;
    XtSetValues(mainMenu,args,nargs);
    
    nargs=0;
    w=XmCreatePushButton(w1,"Overview",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,helpCBWrapper,(XtPointer)0);
    
    nargs=0;
    w=XmCreatePushButton(w1,"Version",args,nargs);
    XtManageChild(w);
    XtAddCallback(w,XmNactivateCallback,helpCBWrapper,(XtPointer)1);
    
  /* Define form for rest of main window */
    nargs=0;
  // KE: Can't seem to do this with resources
    XtSetArg(args[nargs],XmNmarginWidth,MARGIN_WIDTH); nargs++;
    XtSetArg(args[nargs],XmNmarginHeight,MARGIN_HEIGHT); nargs++;
    XtSetArg(args[nargs],XmNwidth,HIST_DEFAULT_WIDTH+2*MARGIN_WIDTH); nargs++;
    XtSetArg(args[nargs],XmNheight,HIST_DEFAULT_HEIGHT+2*MARGIN_HEIGHT); nargs++;
    mainForm=XmCreateForm(mainWindow,"plotForm",args,nargs);
    XtManageChild(mainForm);
    
  // Pop it up (Needs to be done before creating SciPlot widget)
    XtPopup(shell,XtGrabNone); 
}

Widget CHist::createPlot(Widget parent)
{
    int fgcolorid, bgcolorid;
    Widget w;
    
  // Create the widget
    nargs = 0;
    XtSetArg(args[nargs],XmNtopAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNbottomAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNleftAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNrightAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNx,PLOT_X); nargs++;
    XtSetArg(args[nargs],XmNx,PLOT_X); nargs++;
    XtSetArg(args[nargs],XmNy,PLOT_Y); nargs++;
    XtSetArg(args[nargs],XmNwidth,HIST_DEFAULT_WIDTH); nargs++;
    XtSetArg(args[nargs],XmNheight,HIST_DEFAULT_HEIGHT); nargs++;
    XtSetArg(args[nargs],XmNborderWidth,0); nargs++;
    XtSetArg(args[nargs],XmNshadowThickness,2); nargs++;
    XtSetArg(args[nargs],XmNshadowType,XmSHADOW_IN); nargs++;
    XtSetArg(args[nargs],XmNhighlightThickness,0); nargs++;
    if(*pvName) {
	XtSetArg(args[nargs],XtNplotTitle,pvName); nargs++;
	XtSetArg(args[nargs],XtNshowTitle,True); nargs++;
    } else {
	XtSetArg(args[nargs],XtNshowTitle,False); nargs++;
    }
    if(1) {
	XtSetArg(args[nargs],XtNxLabel,"Values"); nargs++;
	XtSetArg(args[nargs],XtNshowXLabel,True); nargs++;
    } else {
	XtSetArg(args[nargs],XtNshowXLabel,False); nargs++;
    }
    if(1) {
	XtSetArg(args[nargs],XtNyLabel,"Counts"); nargs++;
	XtSetArg(args[nargs],XtNshowYLabel,True); nargs++;
    } else {
	XtSetArg(args[nargs],XtNshowYLabel,False); nargs++;
    }
#if 0
  // Set fonts
    XtSetArg(args[nargs],XtNtitleFont,XtFONT_HELVETICA|
      MAX(TITLE_FONT_HEIGHT,MIN_FONT_HEIGHT)); nargs++;
    XtSetArg(args[nargs],XtNaxisFont,XtFONT_HELVETICA|
      MAX(AXIS_FONT_HEIGHT,MIN_FONT_HEIGHT)); nargs++;
    XtSetArg(args[nargs],XtNlabelFont,XtFONT_HELVETICA|
      MAX(LABEL_FONT_HEIGHT,MIN_FONT_HEIGHT)); nargs++;
    XtSetArg(args[nargs],XtNtitleMargin,5); nargs++;
#else
  // Set fonts
    XtSetArg(args[nargs],XtNtitleFont,XtFONT_HELVETICA|TITLE_FONT_HEIGHT); nargs++;
    XtSetArg(args[nargs],XtNaxisFont,XtFONT_HELVETICA|AXIS_FONT_HEIGHT); nargs++;
    XtSetArg(args[nargs],XtNlabelFont,XtFONT_HELVETICA|LABEL_FONT_HEIGHT); nargs++;
#endif

    XtSetArg(args[nargs],XtNshowLegend,False); nargs++;
    XtSetArg(args[nargs],XtNdrawMajor,False); nargs++;
    XtSetArg(args[nargs],XtNdrawMinor,False); nargs++;
    XtSetArg(args[nargs],XtNdragX,False); nargs++;
    XtSetArg(args[nargs],XtNdragY,False); nargs++;
#if 0
  // KE: These extend the origin to include z.  They don't make the
  // axes cross at zero
    XtSetArg(args[nargs],XtNxOrigin,True); nargs++;
    XtSetArg(args[nargs],XtNyOrigin,True); nargs++;
#endif    

  // Y Axis Style
    switch(yAxisStyle) {
    case LOG10_AXIS:
	XtSetArg(args[nargs],XtNyLog,True); nargs++;
	break;
    case LINEAR_AXIS:
    default:
	XtSetArg(args[nargs],XtNyLog,False); nargs++;
	break;
    }

  // Add this pointer as userData
    XtSetArg(args[nargs],XmNuserData,(XtPointer)this); nargs++;

  // Set miscellaneous args
    XtSetArg(args[nargs],XmNtraversalOn,False); nargs++;

  // Create the SciPlot widget
    w = XtCreateWidget("plotWidget", sciplotWidgetClass,
      parent,args,nargs);

  // Have to realize the widget before setting the rest
    XtRealizeWidget(w);

  // Set foreground and background colors
    fgcolorid=SciPlotStoreAllocatedColor(w,fgColor);
    bgcolorid=SciPlotStoreAllocatedColor(w,bgColor);
    SciPlotSetForegroundColor(w, fgcolorid);
    SciPlotSetBackgroundColor(w, bgcolorid);

    switch(axisMode) {
    case AUTO_SCALE_RANGE:
	SciPlotSetXAutoScale(w);
	break;
    case USER_SPECIFIED_RANGE:
	SciPlotSetXUserScale(w, axisXMin, axisXMax);
	break;
    default:
	break;
    }

  // Set Y1 Axis Range
    switch(yRangeStyle) {
    case AUTO_SCALE_RANGE:
	SciPlotSetYAutoScale(w);
#if 0
	SciPlotSetYUserMin(w,0.0);
#endif
	break;
    case USER_SPECIFIED_RANGE:
	SciPlotSetYUserScale(w,minY,maxY);
	break;
    default:
	break;
    }

  // Create the array
    if(xArray && yArray && nBins) {
	plotID=SciPlotListCreateDouble(w,nPoints,xArray,yArray,"Histogram");
      // Change the symbol
#if 1
	SciPlotListSetStyle(w,plotID,1,XtMARKER_NONE,1,XtLINE_SOLID);
#else
	SciPlotListSetStyle(w,plotID,1,XtMARKER_DOT,1,XtLINE_SOLID);
#endif
    } else {
	ERRMSG("%s: No data is defined",
	  pvName);
    }

    SciPlotUpdate(w);
    return w;
}

// Change the X axis
void CHist::updateXAxis(void)
{
    if(!plotW) return;

    switch(axisMode) {
    case AUTO_SCALE_RANGE:
	SciPlotSetXAutoScale(plotW);
	break;
    case USER_SPECIFIED_RANGE:
	SciPlotSetXUserScale(plotW, axisXMin, axisXMax);
	break;
    default:
	break;
    }
    needsUpdate=1;
}


// Sets up the x and y arrays for the bins
void CHist::initializeBins(void)
{
    int i;

  // Determine the upper and lower limits based on the mode
    switch(binMode) {
    case CHANNEL_RANGE:
	binMin=lopr;
	binMax=hopr;
	break;
    case USER_SPECIFIED_RANGE:
	binMin=binUserMin;
	binMax=binUserMax;
	break;
    }

  // Delete any existing arrays
    if(xArray) delete [] xArray;
    if(yArray) delete [] yArray;
 
  // Determine bin parameters   
    nPoints=2*nBins+2;
    binTotalWidth=(binMax-binMin);
    binWidth=binTotalWidth/(double)nBins;
    if(binWidth < 0.0) {
	ERRMSG("%s: Invalid bin width (%g,%g)",
	  pvName,binMin,binMax);
	return;
    }

  // X array
    xArray=new double[nPoints];
    if(!xArray) {
	ERRMSG("%s: No space for x array",pvName);
    }
    for(i=0; i < nBins; i++) {
	xArray[2*i]=binMin+i*binWidth;
	xArray[2*i+1]=binMin+i*binWidth;
    }
    xArray[nPoints-2]=xArray[nPoints-1]=binMax;

  // Y array
    yArray=new double[nPoints];
    if(!yArray) {
	ERRMSG("%s: No space for y array",pvName);
	delete [] xArray;
    }
    for(i=0; i < nPoints; i++) yArray[i]=0.0;

  // Mark for update
    needsUpdate=1;
}

// Puts val in the appropriate bin
void CHist::setBin(double val)
{
    int bin;

  // Check for no bins or out of range value
    if(!xArray || val < binMin || val > binMax) {
	return;
    }

  // Calculate the bin
    bin=(int)((val-binMin)/binWidth)+1;
    if(bin > nBins) bin=nBins;

    bin*=2;

  // Increment the bin
    yArray[bin-1]+=1.0;
    yArray[bin]+=1.0;

    needsUpdate=1;
}

void CHist::startNewName(void)
{
}

// Updates the SciPlot graph from xArray and yArray
void CHist::updatePlot()
{
    if(!needsUpdate || !xArray) return;
    SciPlotListUpdateDouble(plotW,plotID,nPoints,xArray,yArray);
    SciPlotUpdate(plotW);
}

// Default error message.  Can be replaced by the user.
int CHist::errMsg(const char *fmt, ...)
{
    static char string[2048];
    va_list vargs;
    int nchars=0;

    va_start(vargs,fmt);
    nchars+=vsprintf(string,fmt,vargs);
    va_end(vargs);

#ifdef WIN32
  //    MessageBox(NULL,string,"CaObject Error",MB_OK|MB_ICONERROR);
#else
    fprintf(stderr,"%s\n",string);
#endif

    return nchars;
}

// Callback for X/Motif controls for this object
void CHist::mainCB(Widget w, XtPointer cd, XtPointer cbs)
{
    int buttonNumber = (int)cd;
    CHist *hist;

    UNREFERENCED(cbs);
    
    XtVaGetValues(w,XmNuserData,&hist,NULL);

    switch(buttonNumber) {
    case FILE_SAVE_BTN:
	if(!hist->plotW) {
	    XBell(display,50);
	    return;
	}
	if(hist) {
	    const char *fileName=hist->getFileName();

	    if(!fileName || !*fileName) {
		ERRMSG("There is no previously-opened file to save.\n"
		  "Use one of the \"Save As\" options first.");
		return;
	    } else {
	      // Write without checking overwrite
		switch(hist->fileType) {
		case SDDS_File:
		    hist->writeSDDSFile(fileName,False);
		    break;
		case EPS_File:
		    hist->writeEPSFile(fileName,False);
		    break;
		}
	    }
	}
	break;
    case FILE_SAVEAS_SDDS_BTN:
	if(!hist->plotW) {
	    XBell(display,50);
	    return;
	}
	fsdButton=FILE_SAVEAS_SDDS_BTN;
	hist->executeFSD(hist,SDDS_File);
	break;
    case FILE_SAVEAS_EPS_BTN:
	if(!hist->plotW) {
	    XBell(display,50);
	    return;
	}
	fsdButton=FILE_SAVEAS_EPS_BTN;
	hist->executeFSD(hist,EPS_File);
	break;
    case TOOLS_START_BTN:
	start(w,hist,cbs);
	break;
    case TOOLS_STOP_BTN:
	stop(w,hist,cbs);
	break;
    case TOOLS_RESET_BTN:
	reset(w,hist,cbs);
	break;
    case TOOLS_OPTIONS_BTN:
	createOptionsDialog(hist,hist?False:True);
	break;
    case OPTIONS_APPLY_BTN:
	if(optS) {
	    updateOptionsValues(hist,hist?False:True);
	}
	break;
    case OPTIONS_OK_BTN:
	if(optS) {
	    updateOptionsValues(hist,hist?False:True);
	    XtPopdown(optS);
	}
	break;
    case OPTIONS_CANCEL_BTN:
	if(optS) XtPopdown(optS);
	break;
    case OPTIONS_SAVE_BTN:
	if(optS) {
	  // Always defaults (True)
	    updateOptionsValues(hist,True);
	}
	break;
    case OPTIONS_HELP_BTN:
      // Call help with button 0
	if(helpCB) (*helpCB)(w,0,cbs);
	break;
    case QUIT_BTN:
	quit(w,hist,cbs);
	break;
    }
}

void CHist::executeFSD(CHist *hist, FileType fileTypeIn)
{
    XmString histDirMask;
    
    if(!hist) {
	errMsg("executeFSD: Invalid CHist");
	return;
    }
    
  // Determine the suggested extension
    switch(fileTypeIn) {
    case SDDS_File:
	histDirMask=XmStringCreateLocalized(HIST_DIR_SDDS_MASK);
	break;
    case EPS_File:
	histDirMask=XmStringCreateLocalized(HIST_DIR_EPS_MASK);
	break;
    }
	
    if(fileSD == NULL) {
	nargs=0;
	XtSetArg(args[nargs],XmNdialogStyle,
	  XmDIALOG_FULL_APPLICATION_MODAL); nargs++;
	XtSetArg(args[nargs],XmNdirMask,histDirMask); nargs++;
	fileSD=XmCreateFileSelectionDialog(appMainWindow,"fileSD",args,nargs);
	XtAddCallback(fileSD,XmNokCallback,fileSelectCB,(XtPointer)hist);
	XtAddCallback(fileSD,XmNcancelCallback,fileSelectCB,(XtPointer)hist);
	XtUnmanageChild(XmFileSelectionBoxGetChild(fileSD,
	  XmDIALOG_HELP_BUTTON));
	XtManageChild(fileSD);
	XmStringFree(histDirMask);
    } else {
	nargs=0;
	XtSetArg(args[nargs],XmNdirMask,histDirMask); nargs++;
	XtSetValues(fileSD,args,nargs);
	XtRemoveAllCallbacks(fileSD,XmNokCallback);
	XtRemoveAllCallbacks(fileSD,XmNcancelCallback);
	XtAddCallback(fileSD,XmNokCallback,fileSelectCB,(XtPointer)hist);
	XtAddCallback(fileSD,XmNcancelCallback,fileSelectCB,(XtPointer)hist);
	XtManageChild(fileSD);
    }
}

void CHist::fileSelectCB(Widget w, XtPointer clientData, XtPointer callData)
{
    CHist *hist=(CHist *)clientData;
    XmFileSelectionBoxCallbackStruct *cbs =
      (XmFileSelectionBoxCallbackStruct *)callData;

    switch(cbs->reason){
    case XmCR_CANCEL:
	XtUnmanageChild(w);
	break;
    case XmCR_OK:
	if(cbs->value != NULL) {
	    char *fullPathName;
	    int status;
	    
	    XmStringGetLtoR(cbs->value,XmFONTLIST_DEFAULT_TAG,&fullPathName);

	  // Do the appropriate action
	    XFlush(display);
	    switch(fsdButton) {
	    case FILE_SAVEAS_SDDS_BTN:
		status=hist->writeSDDSFile(fullPathName);
		break;
	    case FILE_SAVEAS_EPS_BTN:
		status=hist->writeEPSFile(fullPathName);
		break;
	    }
	    XtFree(fullPathName);
	    XtUnmanageChild(w);
	} else {
	}
	break;
    default:
	break;
    }
}

// Callback for options bin controls
void CHist::optionsBinCB(Widget w, XtPointer cd, XtPointer cbs)
{
    int type = (int)cd;
    int buttonNumber;
  //    CHist *hist;

    UNREFERENCED(cbs);

  // If the type is less than 3, the callback comes from an option
  // menu button.  Find the real type from the userData of the RC
  // parent of the button
    if(type < 3) {
	buttonNumber=type;
	XtVaGetValues(XtParent(w), XmNuserData, &type,NULL);
    }

    switch(type) {
    case OPTIONS_BIN_MODE_BTN:
	break;
    case OPTIONS_BIN_LOW_ENTRY:
	break;
    case OPTIONS_BIN_HIGH_ENTRY:
	break;
    case OPTIONS_BIN_UPDATE_PERIOD_ENTRY:
	break;
    case OPTIONS_BIN_NBINS_ENTRY:
	break;
    }
}

// Callback for options axis controls
void CHist::optionsAxisCB(Widget w, XtPointer cd, XtPointer cbs)
{
    int type = (int)cd;
    int buttonNumber;
  //    CHist *hist;

    UNREFERENCED(cbs);

  // If the type is less than 3, the callback comes from an option
  // menu button.  Find the real type from the userData of the RC
  // parent of the button
    if(type < 3) {
	buttonNumber=type;
	XtVaGetValues(XtParent(w), XmNuserData, &type,NULL);
    }

    switch(type) {
    case OPTIONS_AXIS_MODE_BTN:
	break;
    case OPTIONS_AXIS_MIN_ENTRY:
	break;
    case OPTIONS_AXIS_MAX_ENTRY:
	break;
    }
}

// Callback wrapper for X/Motif help controls.  The user sets his own
// function to do the work.
void CHist::helpCBWrapper(Widget w, XtPointer cd, XtPointer cbs)
{
    if(helpCB) (*helpCB)(w,cd,cbs);
}

void CHist::start(Widget w, XtPointer clientData, XtPointer callData)
{
    CHist *hist=(CHist *)clientData;

    UNREFERENCED(w);
    UNREFERENCED(callData);
    
    if(!hist) {
	errMsg("CHist::start: Invalid CHist");
	return;
    }
    if(hist->histTimerId) {
	XBell(display,50);
	return;
    }

  // Install the timer proc
    hist->histTimerId=XtAppAddTimeOut(appContext,hist->binUpdatePeriod,
      ::histTimerProc,hist);
}

void CHist::stop(Widget w, XtPointer clientData, XtPointer callData)
{
    CHist *hist=(CHist *)clientData;

    UNREFERENCED(w);
    UNREFERENCED(callData);

    if(!hist) {
	errMsg("CHist::stop: Invalid CHist");
	return;
    }
    if(!hist->histTimerId) {
	XBell(display,50);
	return;
    }

  // Uninstall the timer proc
    XtRemoveTimeOut(hist->histTimerId);
    hist->histTimerId=0;
}

void CHist::reset(Widget w, XtPointer clientData, XtPointer callData)
{
    CHist *hist=(CHist *)clientData;

    UNREFERENCED(w);
    UNREFERENCED(callData);

    if(!hist) {
	errMsg("CHist::reset: Invalid CHist");
	return;
    }
    
  // Reset the values
    for(int i=0; i < hist->nPoints; i++) hist->yArray[i]=0.0;

  // Mark for update
    hist->needsUpdate=1;
    hist->updatePlot();
}

// The options dialog should only be popped up from this routine so it
// will be reset correctly
void CHist::createOptionsDialog(CHist *hist, Boolean defaults)
{
    Widget w1,wparent,wrc,form;
    XmString label,opt1,opt2;

  // If the shell is created, just reset it and pop it up
    if(optS) {
	if(!defaults) {
	    if(hist) {
		updateOptionsDialog(hist,False);
	    } else {
		errMsg("CHist::createOptionsDialog: No CHist");
		return;
	    }
	} else {
	    updateOptionsDialog(NULL,True);
	}
	XtPopup(optS,XtGrabNone); 
	return;
    }

  // Create shell if not created
    nargs=0;

  // The following should turn all decorations on except resize handles
    XtSetArg(args[nargs],XmNmwmDecorations,
      MWM_DECOR_ALL|MWM_DECOR_RESIZEH); nargs++;
    XtSetArg(args[nargs],XmNallowShellResize,TRUE); nargs++;

  // This should not be necessary but it fixes Exceed problem
    XtSetArg(args[nargs],XmNmwmFunctions,
      MWM_FUNC_ALL); nargs++;

  // Map window manager menu Close function to nothing for now
    XtSetArg(args[nargs],XmNdeleteResponse,XmDO_NOTHING); nargs++;
    XtSetArg(args[nargs],XmNtitle,"Histogram Options"); nargs++;
    optS = XtCreatePopupShell("optPopup",topLevelShellWidgetClass,
      appMainWindow,args,nargs);
    
  // Fix up WM close
    XmAddWMProtocolCallback(optS,WM_DELETE_WINDOW,
      mainCB,(XtPointer)OPTIONS_CANCEL_BTN);

  // Create a form
    nargs=0;
    form=XmCreateForm(optS,"optForm",args,nargs);

  /* PV name */
    nargs=0;
    XtSetArg(args[nargs],XmNtopAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNrightAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNleftAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNalignment,XmALIGNMENT_CENTER); nargs++;
    w1=XmCreateLabel(form,"PVName",args,nargs);
    XtManageChild(w1);
    optNameW=w1;
    wparent=w1;

  // Separator
    nargs=0;
    XtSetArg(args[nargs],XmNtopAttachment,XmATTACH_WIDGET); nargs++;
    XtSetArg(args[nargs],XmNtopWidget,wparent); nargs++;
    XtSetArg(args[nargs],XmNrightAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNleftAttachment,XmATTACH_FORM); nargs++;
    wparent=XmCreateSeparator(form,"separator",args,nargs);
    XtManageChild(wparent);

  /* Bin parameters */
    nargs=0;
    XtSetArg(args[nargs],XmNtopAttachment,XmATTACH_WIDGET); nargs++;
    XtSetArg(args[nargs],XmNtopWidget,wparent); nargs++;
    XtSetArg(args[nargs],XmNrightAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNleftAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNorientation,XmVERTICAL); nargs++;
    wparent=XmCreateRowColumn(form,"rowCol",args,nargs);

    nargs=0;
    w1=XmCreateLabel(wparent,"Bin Parameters",args,nargs);
    XtManageChild(w1);

    label=XmStringCreateLocalized("Limits Mode:");
    opt1=XmStringCreateLocalized("Channel");
    opt2=XmStringCreateLocalized("User Specified");
    w1=XmVaCreateSimpleOptionMenu(wparent,"optionMenu",
      label,'\0',0,optionsBinCB,
      XmVaPUSHBUTTON,opt1,'\0',NULL,NULL,
      XmVaPUSHBUTTON,opt2,'\0',NULL,NULL,
      XmNmarginWidth,OPTIONS_INDENT,
      XmNuserData,OPTIONS_BIN_MODE_BTN,
      NULL);
    XtManageChild(w1);
    XmStringFree(label);
    XmStringFree(opt1);
    XmStringFree(opt2);
    optBinModeW=w1;

  // Row column 1 for buttons
    nargs=0;
    XtSetArg(args[nargs],XmNorientation,XmHORIZONTAL); nargs++;
    wrc=XmCreateRowColumn(wparent,"rowCol",args,nargs);

    createLabelBox(wrc,"labelBox","Channel Lower Limit:",10,"",optionsBinCB,
      (XtPointer)OPTIONS_BIN_LOW_ENTRY,&optBinLoprE);

    createLabelBox(wrc,"labelBox","Channel Upper Limit:",10,"",optionsBinCB,
      (XtPointer)OPTIONS_BIN_HIGH_ENTRY,&optBinHoprE);

    nargs=0;
    XtSetArg(args[nargs],XmNeditable,False); nargs++;
    XtSetArg(args[nargs],XmNcursorPositionVisible,False); nargs++;
    XtSetValues(optBinLoprE, args,nargs);
    XtSetValues(optBinHoprE, args,nargs);

    XtManageChild(wrc);

  // Row column 2 for buttons
    nargs=0;
    XtSetArg(args[nargs],XmNorientation,XmHORIZONTAL); nargs++;
    wrc=XmCreateRowColumn(wparent,"rowCol",args,nargs);

    createLabelBox(wrc,"labelBox","User-Specified Lower Limit:",10,"",optionsBinCB,
      (XtPointer)OPTIONS_BIN_LOW_ENTRY,&optBinLowE);

    createLabelBox(wrc,"labelBox","User-Specified Upper Limit:",10,"",optionsBinCB,
      (XtPointer)OPTIONS_BIN_HIGH_ENTRY,&optBinHighE);

    XtManageChild(wrc);

  // Row column 3 for buttons
    nargs=0;
    XtSetArg(args[nargs],XmNorientation,XmHORIZONTAL); nargs++;
    wrc=XmCreateRowColumn(wparent,"rowCol",args,nargs);

    createLabelBox(wrc,"labelBox","Number of bins:",10,"",optionsBinCB,
      (XtPointer)OPTIONS_BIN_NBINS_ENTRY,&optBinNBinsE);

    createLabelBox(wrc,"labelBox","Update Period (ms):",10,"",optionsBinCB,
      (XtPointer)OPTIONS_BIN_UPDATE_PERIOD_ENTRY,&optBinUpdatePeriodE);

    XtManageChild(wrc);

  // Manage the parent
    XtManageChild(wparent);

  // Separator
    nargs=0;
    XtSetArg(args[nargs],XmNtopAttachment,XmATTACH_WIDGET); nargs++;
    XtSetArg(args[nargs],XmNtopWidget,wparent); nargs++;
    XtSetArg(args[nargs],XmNrightAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNleftAttachment,XmATTACH_FORM); nargs++;
    wparent=XmCreateSeparator(form,"separator",args,nargs);
    XtManageChild(wparent);

  /* Axis limits */
    nargs=0;
    XtSetArg(args[nargs],XmNtopAttachment,XmATTACH_WIDGET); nargs++;
    XtSetArg(args[nargs],XmNtopWidget,wparent); nargs++;
    XtSetArg(args[nargs],XmNrightAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNleftAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNorientation,XmVERTICAL); nargs++;
    wparent=XmCreateRowColumn(form,"rowCol",args,nargs);

    nargs=0;
    w1=XmCreateLabel(wparent,"X Axis Limits",args,nargs);
    XtManageChild(w1);

    label=XmStringCreateLocalized("Axis Mode:");
    opt1=XmStringCreateLocalized("Autoscale");
    opt2=XmStringCreateLocalized("User Specified");
    w1=XmVaCreateSimpleOptionMenu(wparent,"optionMenu",
      label,'\0',0,optionsAxisCB,
      XmVaPUSHBUTTON,opt1,'\0',NULL,NULL,
      XmVaPUSHBUTTON,opt2,'\0',NULL,NULL,
      XmNmarginWidth,OPTIONS_INDENT,
      XmNuserData,OPTIONS_BIN_MODE_BTN,
      NULL);
    XtManageChild(w1);
    XmStringFree(label);
    XmStringFree(opt1);
    XmStringFree(opt2);
    optAxisModeW=w1;

  // Row column for buttons
    nargs=0;
    XtSetArg(args[nargs],XmNorientation,XmHORIZONTAL); nargs++;
    wrc=XmCreateRowColumn(wparent,"rowCol",args,nargs);

    createLabelBox(wrc,"labelBox","Lower Limit:",10,"",optionsAxisCB,
      (XtPointer)OPTIONS_AXIS_MIN_ENTRY,&optAxisMinE);

    createLabelBox(wrc,"labelBox","Upper Limit:",10,"",optionsAxisCB,
      (XtPointer)OPTIONS_AXIS_MIN_ENTRY,&optAxisMaxE);

    XtManageChild(wrc);

  // Manage the parent
    XtManageChild(wparent);

  // Separator
    nargs=0;
    XtSetArg(args[nargs],XmNtopAttachment,XmATTACH_WIDGET); nargs++;
    XtSetArg(args[nargs],XmNtopWidget,wparent); nargs++;
    XtSetArg(args[nargs],XmNrightAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNleftAttachment,XmATTACH_FORM); nargs++;
    wparent=XmCreateSeparator(form,"separator",args,nargs);
    XtManageChild(wparent);

  // Action area buttons
    nargs=0;
    XtSetArg(args[nargs],XmNtopAttachment,XmATTACH_WIDGET); nargs++;
    XtSetArg(args[nargs],XmNtopWidget,wparent); nargs++;
    XtSetArg(args[nargs],XmNrightAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNleftAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNbottomAttachment,XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNshadowThickness,0); nargs++;
    XtSetArg(args[nargs],XmNfractionBase,80); nargs++;
    XtSetArg(args[nargs],XmNskipAdjust,True); nargs++;
    wparent = XmCreateForm(form,"actionArea",args,nargs);

    nargs=0;
#if 0
    XtSetArg(args[nargs],XmNuserData,hist); nargs++;
#endif
    XtSetArg(args[nargs],XmNtopAttachment, XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNbottomAttachment, XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNleftAttachment, XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNleftPosition,5); nargs++;
    XtSetArg(args[nargs],XmNrightAttachment, XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNrightPosition,15); nargs++;
    w1=XmCreatePushButton(wparent,"Apply",args,nargs);
    XtAddCallback(w1,XmNactivateCallback,mainCB,(XtPointer)OPTIONS_APPLY_BTN);
    XtManageChild(w1);
    optionsApplyB=w1;
    
    nargs=0;
#if 0
    XtSetArg(args[nargs],XmNuserData,hist); nargs++;
#endif
    XtSetArg(args[nargs],XmNtopAttachment, XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNbottomAttachment, XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNleftAttachment, XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNleftPosition,20); nargs++;
    XtSetArg(args[nargs],XmNrightAttachment, XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNrightPosition,30); nargs++;
    w1=XmCreatePushButton(wparent,"OK",args,nargs);
    XtAddCallback(w1,XmNactivateCallback,mainCB,(XtPointer)OPTIONS_OK_BTN);
    XtManageChild(w1);
    optionsOkB=w1;
    
    nargs=0;
#if 0
    XtSetArg(args[nargs],XmNuserData,hist); nargs++;
#endif
    XtSetArg(args[nargs],XmNtopAttachment, XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNbottomAttachment, XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNleftAttachment, XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNleftPosition,35); nargs++;
    XtSetArg(args[nargs],XmNrightAttachment, XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNrightPosition,45); nargs++;
    w1=XmCreatePushButton(wparent,"Cancel",args,nargs);
    XtAddCallback(w1,XmNactivateCallback,mainCB,(XtPointer)OPTIONS_CANCEL_BTN);
    XtManageChild(w1);
    optionsCancelB=w1;
    
    nargs=0;
#if 0
    XtSetArg(args[nargs],XmNuserData,hist); nargs++;
#endif
    XtSetArg(args[nargs],XmNtopAttachment, XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNbottomAttachment, XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNleftAttachment, XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNleftPosition,50); nargs++;
    XtSetArg(args[nargs],XmNrightAttachment, XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNrightPosition,60); nargs++;
    w1=XmCreatePushButton(wparent,"Save",args,nargs);
    XtAddCallback(w1,XmNactivateCallback,mainCB,(XtPointer)OPTIONS_SAVE_BTN);
    XtManageChild(w1);
    optionsSaveB=w1;
    
    nargs=0;
#if 0
    XtSetArg(args[nargs],XmNuserData,hist); nargs++;
#endif
    XtSetArg(args[nargs],XmNtopAttachment, XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNbottomAttachment, XmATTACH_FORM); nargs++;
    XtSetArg(args[nargs],XmNleftAttachment, XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNleftPosition,65); nargs++;
    XtSetArg(args[nargs],XmNrightAttachment, XmATTACH_POSITION); nargs++;
    XtSetArg(args[nargs],XmNrightPosition,75); nargs++;
    w1=XmCreatePushButton(wparent,"Help",args,nargs);
    XtAddCallback(w1,XmNactivateCallback,mainCB,(XtPointer)OPTIONS_HELP_BTN);
    XtManageChild(w1);
    optionsHelpB=w1;

  // Manage the parent
    XtManageChild(wparent);

  // Manage the form
    XtManageChild(form);

  // Pop it up
    if(!defaults) {
	if(hist) {
	    updateOptionsDialog(hist,False);
	} else {
	    errMsg("CHist::createOptionsDialog: No CHist");
	    return;
	}
    } else {
	updateOptionsDialog(hist,True);
    }
    XtPopup(optS,XtGrabNone); 
}

void CHist::updateOptionsDialog(CHist *hist, Boolean defaults)
{
    char string[1024];     /* Danger: Fixed length */
    XmString xmString;
    int binModeLoc;
    double binUserMinLoc;
    double binUserMaxLoc;
    unsigned binUpdatePeriodLoc;
    int nBinsLoc;
    int axisModeLoc;
    double axisXMinLoc;
    double axisXMaxLoc;

    short precLoc;
    XtPointer userData;

    if(!defaults) {
      // Is a class instance
	if(!hist) {
	    errMsg("CHist::updateOptionsDialog: Invalid CHist");
	    return;
	}
	binModeLoc=hist->binMode;
	binUserMinLoc=hist->binUserMin;
	binUserMaxLoc=hist->binUserMax;
	binUpdatePeriodLoc=hist->binUpdatePeriod;
	nBinsLoc=hist->nBins;
	axisModeLoc=hist->axisMode;
	axisXMinLoc=hist->axisXMin;
	axisXMaxLoc=hist->axisXMax;
	precLoc=hist->prec;
	userData=(XtPointer)hist;
    } else {
	binModeLoc=options.binMode;
	binUserMinLoc=options.binUserMin;
	binUserMaxLoc=options.binUserMax;
	binUpdatePeriodLoc=options.binUpdatePeriod;
	nBinsLoc=options.nBins;
	axisModeLoc=options.axisMode;
	axisXMinLoc=options.axisXMin;
	axisXMaxLoc=options.axisXMax;
	precLoc=6;
	userData=NULL;
    }

  // Set the user data for the buttons
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,userData); nargs++;
    XtSetValues(optS,args,nargs);
    XtSetValues(optionsApplyB,args,nargs);
    XtSetValues(optionsOkB,args,nargs);
    XtSetValues(optionsCancelB,args,nargs);
    XtSetValues(optionsSaveB,args,nargs);
    XtSetValues(optionsHelpB,args,nargs);

  // PV Name
    if(!defaults) {
	xmString=XmStringCreateLocalized(hist->pvName);
    } else {
	xmString=XmStringCreateLocalized("DEFAULTS");
    }
    nargs=0;
    XtSetArg(args[nargs],XmNlabelString,xmString); nargs++;
    XtSetArg(args[nargs],XmNuserData,userData); nargs++;
    XtSetValues(optNameW,args,nargs);
    XmStringFree(xmString);

  // Bin Parameters
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,userData); nargs++;
    XtSetValues(optBinModeW,args,nargs);
    switch(binModeLoc) {
    case CHANNEL_RANGE:
	optionMenuSet(optBinModeW,0);
	break;
    case USER_SPECIFIED_RANGE:
	optionMenuSet(optBinModeW,1);
	break;
    }

  // Bin channel low limit
    if(!defaults) {
	cvtDoubleToString(hist->lopr,string,precLoc);
    } else {
	*string='\0';
    }
    XmTextFieldSetString(optBinLoprE,string);
    
  // Bin channel high limit
    if(!defaults) {
	cvtDoubleToString(hist->hopr,string,precLoc);
    } else {
	*string='\0';
    }
    XmTextFieldSetString(optBinHoprE,string);

  // Bin user low limit
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,userData); nargs++;
    XtSetValues(optBinLowE,args,nargs);
    cvtDoubleToString(binUserMinLoc,string,precLoc);
    XmTextFieldSetString(optBinLowE,string);
    
  // Bin user high limit
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,userData); nargs++;
    XtSetValues(optBinHighE,args,nargs);
    cvtDoubleToString(binUserMaxLoc,string,precLoc);
    XmTextFieldSetString(optBinHighE,string);

  // N bins
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,userData); nargs++;
    XtSetValues(optBinNBinsE,args,nargs);
    sprintf(string,"%d",nBinsLoc);
    XmTextFieldSetString(optBinNBinsE,string);

  // Update period
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,userData); nargs++;
    XtSetValues(optBinUpdatePeriodE,args,nargs);
    sprintf(string,"%u",binUpdatePeriodLoc);
    XmTextFieldSetString(optBinUpdatePeriodE,string);

  // Axis parameters
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,userData); nargs++;
    XtSetValues(optAxisModeW,args,nargs);
    switch(axisModeLoc) {
    case AUTO_SCALE_RANGE:
	optionMenuSet(optAxisModeW,0);
	break;
    case USER_SPECIFIED_RANGE:
	optionMenuSet(optAxisModeW,1);
	break;
    }
    
  // Axis low limit
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,userData); nargs++;
    XtSetValues(optAxisMinE,args,nargs);
    cvtDoubleToString(axisXMinLoc,string,precLoc);
    XmTextFieldSetString(optAxisMinE,string);
    
  // Axis High limit
    nargs=0;
    XtSetArg(args[nargs],XmNuserData,userData); nargs++;
    XtSetValues(optAxisMaxE,args,nargs);
    cvtDoubleToString(axisXMaxLoc,string,precLoc);
    XmTextFieldSetString(optAxisMaxE,string);
}

void CHist::updateOptionsValues(CHist *hist, Boolean defaults)
{
    int binModeNew,binModeOld;
    double binUserMinNew,binUserMinOld;
    double binUserMaxNew,binUserMaxOld;
    unsigned binUpdatePeriodNew,binUpdatePeriodOld;
    int nBinsNew,nBinsOld;
    int axisModeNew,axisModeOld;
    double axisXMinNew,axisXMinOld;
    double axisXMaxNew,axisXMaxOld;
    Widget menuWidget;
    Widget menuHistory;
    WidgetList children;
    Cardinal numChildren;
    char *string;

    if(!defaults) {
      // Is a class instance
	if(!hist) {
	    errMsg("CHist::updateOptionsValues: Invalid CHist");
	    return;
	}

      // Save the old values to look for changes
	binModeOld=hist->binMode;
	binUserMinOld=hist->binUserMin;
	binUserMaxOld=hist->binUserMax;
	binUpdatePeriodOld=hist->binUpdatePeriod;
	nBinsOld=hist->nBins;
	axisModeOld=hist->axisMode;
	axisXMinOld=hist->axisXMin;
	axisXMaxOld=hist->axisXMax;
    }
    
  // Bin Parameters
    nargs = 0;
    XtSetArg(args[nargs],XmNsubMenuId,&menuWidget); nargs++;
    XtSetArg(args[nargs],XmNmenuHistory,&menuHistory); nargs++;
    XtGetValues(optBinModeW,args,nargs);

    nargs = 0;
    XtSetArg(args[nargs],XmNchildren,&children); nargs++;
    XtSetArg(args[nargs],XmNnumChildren,&numChildren); nargs++;
    XtGetValues(menuWidget,args,nargs);

    if(menuHistory == children[0]) {
	binModeNew=CHANNEL_RANGE;
    } else if(menuHistory == children[1]) {
	binModeNew=USER_SPECIFIED_RANGE;
    }

  // Bin user low limit
    string=XmTextFieldGetString(optBinLowE);
    XmTextSetCursorPosition(optBinLowE,0);
    binUserMinNew=atof(string);
    XtFree(string);
    
  // Bin user high limit
    string=XmTextFieldGetString(optBinHighE);
    XmTextSetCursorPosition(optBinHighE,0);
    binUserMaxNew=atof(string);
    XtFree(string);
    
  // N bins
    string=XmTextFieldGetString(optBinNBinsE);
    XmTextSetCursorPosition(optBinNBinsE,0);
    nBinsNew=atoi(string);
    XtFree(string);

  // Update period
    string=XmTextFieldGetString(optBinUpdatePeriodE);
    XmTextSetCursorPosition(optBinUpdatePeriodE,0);
    binUpdatePeriodNew=atol(string);
    XtFree(string);

  // Axis parameters
    nargs = 0;
    XtSetArg(args[nargs],XmNsubMenuId,&menuWidget); nargs++;
    XtSetArg(args[nargs],XmNmenuHistory,&menuHistory); nargs++;
    XtGetValues(optAxisModeW,args,nargs);

    nargs = 0;
    XtSetArg(args[nargs],XmNchildren,&children); nargs++;
    XtSetArg(args[nargs],XmNnumChildren,&numChildren); nargs++;
    XtGetValues(menuWidget,args,nargs);

    if(menuHistory == children[0]) {
	axisModeNew=AUTO_SCALE_RANGE;
    } else if(menuHistory == children[1]) {
	axisModeNew=USER_SPECIFIED_RANGE;
    }
    
  // Axis low limit
    string=XmTextFieldGetString(optAxisMinE);
    XmTextSetCursorPosition(optAxisMinE,0);
    axisXMinNew=atof(string);
    XtFree(string);

  // Axis high limit
    string=XmTextFieldGetString(optAxisMaxE);
    XmTextSetCursorPosition(optAxisMaxE,0);
    axisXMaxNew=atof(string);
    XtFree(string);
    
  // Handle the results
    if(defaults) {
      // Is the defaults
	options.binMode=binModeNew;
	options.binUserMin=binUserMinNew;
	options.binUserMax=binUserMaxNew;
	options.binUpdatePeriod=binUpdatePeriodNew;
	options.nBins=nBinsNew;
	options.axisMode=axisModeNew;
	options.axisXMin=axisXMinNew;
	options.axisXMax=axisXMaxNew;
	options.write();
    } else {
      // Is a class instance
	hist->binMode=binModeNew;
	hist->binUserMin=binUserMinNew;
	hist->binUserMax=binUserMaxNew;
	hist->binUpdatePeriod=binUpdatePeriodNew;
	hist->nBins=nBinsNew;
	hist->axisMode=axisModeNew;
	hist->axisXMin=axisXMinNew;
	hist->axisXMax=axisXMaxNew;
	
      // Decide what to redo
	if(binModeOld != hist->binMode ||
	  binUserMinOld != hist->binUserMin ||
	  binUserMaxOld != hist->binUserMax ||
	  nBinsOld != hist->nBins) {
	    hist->initializeBins();
	}
	if(axisModeOld != hist->axisMode ||
	  axisXMinOld != hist->axisXMin ||
	  axisXMaxOld != hist->axisXMax) {
	    hist->updateXAxis();
	}
	if(binUpdatePeriodOld != hist->binUpdatePeriod && hist->histTimerId	) {
	  // Uninstall the timer proc and reinstall
	    XtRemoveTimeOut(hist->histTimerId);
	    histTimerProc((XtPointer)hist,NULL);
	}
      // Update (Will not be done unless needsUpdate is 1)
	hist->updatePlot();
    }
}

void CHist::quit(Widget w, XtPointer clientData, XtPointer callData)
{
    CHist *hist=(CHist *)clientData;

    UNREFERENCED(w);
    UNREFERENCED(callData);

    if(!hist) {
	errMsg("CHist::quit: Invalid CHist");
	return;
    }

  // Delete the object
    delete hist;
}

// Called by CA when the value of the PV changes.  It changes the
// current stored value.
void CHist::valueChangedCB(CCaObject2 *chanIn)
{
    CHist *hist=(CHist *)chanIn->getUserData();
    if(hist) {
	if(chanIn != hist->chan) {
	    ERRMSG("valueChangedCB: Inconsistent chan's");
	    return;
	}
	hist->value=chanIn->getValue();
#if DEBUG_CA
	printf("CHist::valueChangedCB: value=%g\n",chanIn->getValue());
#endif
    } else {
	ERRMSG("valueChangedCB: Invalid CHist");
    }
}

// Called by CA when the graphical info for the PV changes
void CHist::graphicalInfoCB(CCaObject2 *chanIn)
{
    CHist *hist=(CHist *)chanIn->getUserData();
    if(hist) {
	if(chanIn != hist->chan) {
	    ERRMSG("graphicalInfoCB: Inconsistent chan's");
	    return;
	}
	hist->value=chanIn->getValue();
    } else {
	ERRMSG("graphicalInfoCB: Invalid CHist");
    }


    switch(chanIn->getDataType()) {
    case DBF_STRING :
	ERRMSG("graphicalInfoCB: Channel type is DBF_STRING for %s",
	  hist->pvName);
	return;
    case DBF_ENUM :
    case DBF_CHAR :
    case DBF_INT :
    case DBF_LONG :
    case DBF_FLOAT :
    case DBF_DOUBLE :
	hist->hopr=chanIn->getHopr();
	hist->lopr=chanIn->getLopr();
	hist->value=chanIn->getValue();
	hist->prec=chanIn->getPrecision();
	break;
    default :
	ERRMSG("graphicalInfoCB: Unknown channel type for %s",
	  hist->pvName);
	break;
    }
    if((hist->hopr == 0.0) && (hist->lopr == 0.0)) {
	hist->hopr += 1.0;
    }

  // Initialize the first time
    if(!hist->initialized) {
      // Bins
	hist->initializeBins();
	
      // Create SciPlot widget
	hist->plotW=hist->createPlot(hist->mainForm);
	XtManageChild(hist->plotW);
	
      // Start a timer for the histogram time interval
	start(NULL,hist,NULL);

	hist->initialized=1;
    }
}

// Polls CA on a regular basis so it can do its thing
void CHist::pollTimerProc(XtPointer cd, XtIntervalId *id)
{
    UNREFERENCED(id);

    ca_poll();

  /* Reinstall the timer proc to be called again in POLL_INTERVAL ms */
    pollTimerId=XtAppAddTimeOut(appContext,POLL_INTERVAL,::pollTimerProc,cd);
}

void CHist::histTimerProc(XtPointer cd, XtIntervalId *iid)
{
    UNREFERENCED(iid);
    CHist *hist=(CHist *)cd;

    if(!hist) {
	ERRMSG("histTimerProc: Invalid CHist");
	return;
    }

#if DEBUG_CA
    printf("CHist::histTimerProcCB: value=%g\n",hist->value);
#endif
    hist->setBin(hist->value);
    hist->needsUpdate=1;
    hist->updatePlot();

  /* Reinstall the timer proc to be called again in binUpdatePeriod ms */
    hist->histTimerId=XtAppAddTimeOut(appContext,hist->binUpdatePeriod,
      ::histTimerProc,hist);
}

// The following are wrappers so that CA can use "C" calling
// convention.  They call the static class methods, which do the real
// work.

// KE: This may not be necessary for these CaObject2 functions

static void valueChangedCB(CCaObject2 *chanIn)
{
    CHist::valueChangedCB(chanIn);
}

static void graphicalInfoCB(CCaObject2 *chanIn)
{
    CHist::graphicalInfoCB(chanIn);
}

static void pollTimerProc(XtPointer cd, XtIntervalId *iid)
{
    CHist::pollTimerProc(cd,iid);
}

static void histTimerProc(XtPointer cd, XtIntervalId *iid)
{
    CHist::histTimerProc(cd,iid);
}
